use num::complex::ComplexFloat;
use num::integer::Roots;
use num::traits::real::Real;
use num::{Float, FromPrimitive, Signed};
use realfft::{RealFftPlanner, RealToComplex};
use ringbuf::{HeapRb, Rb};
use rustfft::Fft;
use rustfft::{num_complex::Complex};

use std::iter::{repeat};



use std::{
    fmt::Debug,
    sync::{Arc, Mutex},
};

use std::f64::consts::SQRT_2;

type FloatType = f32;
type SampleType = f32;

pub trait Analyzer {
    /// Process some data, and return whether the analysis was updated as a result
    fn process_data(&mut self, data: &[SampleType]) -> bool;
    fn set_samplerate(&mut self, rate: FloatType);
    fn set_length(&mut self, length: usize);
    fn set_channels(&mut self, channels: usize);
}

#[derive(Debug, Clone, Default)]
pub struct FftAnalysis<T> {
    pub signal_frequency: T,
    pub signal_power: T,
    pub total_power: T,
    pub noise_power: T,
    pub thd_plus_noise: T,
    pub max: T,
    pub min: T,
    pub fft: Vec<(T, T, T)>, // Frequency, Magnitude, Energy
}

struct AudioBuf<T> {
    samples: Vec<HeapRb<T>>,
    read_since_last_update: usize,
}

pub trait WindowFunction<T: Real> {
    fn amplitude_correction_factor(&self) -> T;
    fn energy_correction_factor(&self) -> T;
    fn apply(&self, buf: &mut [T]);
}

struct HanningWindow<T: Real>(Vec<T>);

impl<T: Real + FromPrimitive> HanningWindow<T> {
    fn new(fft_length: usize) -> Self {
        Self(
            apodize::hanning_iter(fft_length)
                .map(|x| T::from_f64(x).unwrap())
                .collect(),
        )
    }
}

impl<T: Real + FromPrimitive> WindowFunction<T> for HanningWindow<T> {
    fn amplitude_correction_factor(&self) -> T {
        T::from_f64(2.0).unwrap()
    }
    fn energy_correction_factor(&self) -> T {
        T::from_f64(2.0 / num::Float::sqrt(3.0 / 2.0)).unwrap()
    }
    fn apply(&self, buf: &mut [T]) {
        for (a, b) in buf.iter_mut().zip(self.0.iter()) {
            *a = *a * *b
        }
    }
}

pub fn lin_to_db(lin: f64) -> f64 {
    20.0 * lin.log10()
}

pub fn lin_to_dbfs(lin: f64) -> f64 {
    20.0 * (lin * SQRT_2).log10()
}

pub struct FftAnalyzer {
    pub last_analysis: Arc<Mutex<Vec<FftAnalysis<SampleType>>>>,

    fft_length: usize,
    sample_rate: FloatType,
    channels: usize,
    norm_factor: FloatType,
    bin_freqs: Vec<FloatType>,

    planner: RealFftPlanner<FloatType>,
    processor: Arc<dyn RealToComplex<FloatType>>,
    window: HanningWindow<FloatType>,

    audio_buf: Arc<Mutex<AudioBuf<FloatType>>>,
}

impl FftAnalyzer {
    pub fn new(fft_length: usize, sample_rate: FloatType, channels: usize) -> Self {
        let mut planner = RealFftPlanner::new();
        let bin_freqs = (0..fft_length)
            .map(|x| {
                FloatType::from_f64(x as f64).unwrap() * sample_rate
                    / FloatType::from_f64(fft_length as f64).unwrap()
            })
            .collect();
        Self {
            last_analysis: Arc::new(Mutex::new(Vec::with_capacity(channels))),
            fft_length,
            sample_rate,
            channels,
            norm_factor: FloatType::from_f64(1.0 / (fft_length as f64).sqrt()).unwrap(),
            bin_freqs,

            processor: planner.plan_fft_forward(fft_length),
            planner,
            window: HanningWindow::new(fft_length),
            audio_buf: Arc::new(Mutex::new(AudioBuf::<SampleType> {
                samples: repeat(0)
                    .take(channels)
                    .map(|_| HeapRb::new(fft_length))
                    .collect(),
                read_since_last_update: 0,
            })),
        }
    }
    fn time_domain_process(&mut self, buf: &Vec<FloatType>) -> (FloatType, FloatType, FloatType) {
        let mut rms_accumulator: FloatType = 0.0;
        let mut local_max: FloatType = 0.0;
        let mut local_min: FloatType = 0.0;

        for sample in buf {
            rms_accumulator += sample * sample;
            if *sample > local_max {
                local_max = *sample
            }
            if *sample < local_min {
                local_min = *sample
            }
        }

        let total_power = (rms_accumulator / buf.len() as f32).sqrt();

        (local_min, local_max, total_power)
    }
    fn process_frame(&mut self) {
        let _channels_float = FloatType::from(self.channels as f32);
        let mut result = Vec::with_capacity(self.channels);

        for chan in 0..self.channels {
            let mut in_buf: Vec<FloatType> = self.audio_buf.lock().unwrap().samples[chan]
                .iter()
                .map(|x| *x)
                .collect();

            let td_result = self.time_domain_process(&in_buf);

            // apply window
            self.window.apply(&mut in_buf);

            let mut out_buf =
                Vec::from_iter(repeat(Complex { re: 0.0, im: 0.0 }).take(self.fft_length / 2 + 1));

            // do fft
            self.processor
                .process(&mut in_buf, out_buf.as_mut_slice())
                .expect("fft failed?");

            let analysis_buf = out_buf
                .iter()
                .enumerate()
                .map(|(bin, complex)| {
                    // get absolute value and normalize based on length
                    let raw_mag = (complex
                        / FloatType::from_f64(self.fft_length as f64 / 2.0).unwrap())
                    .abs();
                    (
                        self.bin_freqs[bin],
                        raw_mag * self.window.amplitude_correction_factor(),
                        raw_mag * self.window.energy_correction_factor(),
                    )
                })
                .collect();
            result.push(FftAnalysis {
                signal_frequency: 0.0,
                signal_power: 0.0,
                total_power: td_result.2,
                noise_power: 0.0,
                thd_plus_noise: 0.0,
                max: td_result.1,
                min: td_result.0,
                fft: analysis_buf,
            });
        }

        let mut lock = self.last_analysis.lock().unwrap();
        *lock = result;
    }
    fn progress(&self) -> f32 {
        self.audio_buf.lock().unwrap().read_since_last_update as f32 / self.fft_length as f32
    }
}

impl Analyzer for FftAnalyzer {
    fn process_data(&mut self, data: &[SampleType]) -> bool {
        let _channels_float = FloatType::from(self.channels as f32);

        let mut buf = self.audio_buf.lock().unwrap();
        for chan in 0..self.channels {
            buf.samples[chan].push_iter_overwrite(data.iter().skip(chan).step_by(self.channels).map(|x| *x));
        }
        buf.read_since_last_update += data.len() / self.channels;

        if buf.samples[0].len() >= self.fft_length && buf.read_since_last_update >= self.fft_length / 4 {
            buf.read_since_last_update = 0;
            drop(buf);
            self.process_frame();
            true
        } else {
            false
        }
    }
    fn set_channels(&mut self, channels: usize) {
        self.channels = channels;
    }
    fn set_length(&mut self, _length: usize) {
        unimplemented!(); // Need to rebuild the window and plan
    }
    fn set_samplerate(&mut self, rate: FloatType) {
        self.sample_rate = rate;
    }
}
